@@ -75,13 +75,34 @@ _GJS_CAIRO_CONTEXT_DEFINE_FUNC_BEGIN(method)                               \
     JS_SET_RVAL(context, vp, BOOLEAN_TO_JSVAL(ret));                       \
 _GJS_CAIRO_CONTEXT_DEFINE_FUNC_END
 
+#define _GJS_CAIRO_CONTEXT_DEFINE_FUNC2FFAFF(method, cfunc, n1, n2)        \
+_GJS_CAIRO_CONTEXT_DEFINE_FUNC_BEGIN(method)                               \
+    double arg1, arg2;                                                     \
+    if (!gjs_parse_args(context, #method, "ff", argc, JS_ARGV(context, vp), \
+                        #n1, &arg1, #n2, &arg2))                           \
+        return JS_FALSE;                                                   \
+    cr = gjs_cairo_context_get_context(context, obj);                      \
+    cfunc(cr, &arg1, &arg2);                                               \
+    if (cairo_status(cr) == CAIRO_STATUS_SUCCESS) {                        \
+      JSObject *array = JS_NewArrayObject(context, 0, NULL);               \
+      if (!array)                                                          \
+        return JS_FALSE;                                                   \
+      jsval r;                                                             \
+      if (!JS_NewNumberValue(context, arg1, &r)) return JS_FALSE;          \
+      if (!JS_SetElement(context, array, 0, &r)) return JS_FALSE;          \
+      if (!JS_NewNumberValue(context, arg2, &r)) return JS_FALSE;          \
+      if (!JS_SetElement(context, array, 1, &r)) return JS_FALSE;          \
+      JS_SET_RVAL(context, vp, OBJECT_TO_JSVAL(array));                    \
+    }                                                                      \
+_GJS_CAIRO_CONTEXT_DEFINE_FUNC_END
+
 #define _GJS_CAIRO_CONTEXT_DEFINE_FUNC0AFF(method, cfunc)                  \
 _GJS_CAIRO_CONTEXT_DEFINE_FUNC_BEGIN(method)                               \
     double arg1, arg2;                                                     \
    _GJS_CAIRO_CONTEXT_CHECK_NO_ARGS(method)                                \
     cr = gjs_cairo_context_get_context(context, obj);                      \
     cfunc(cr, &arg1, &arg2);                                               \
-    {                                                                      \
+    if (cairo_status(cr) == CAIRO_STATUS_SUCCESS) {                        \
       JSObject *array = JS_NewArrayObject(context, 0, NULL);               \
       if (!array)                                                          \
         return JS_FALSE;                                                   \
@@ -231,7 +252,7 @@ typedef struct {
     cairo_t * cr;
 } GjsCairoContext;
 
-GJS_DEFINE_PROTO("CairoContext", gjs_cairo_context)
+GJS_DEFINE_PROTO("CairoContext", cairo_context)
 GJS_DEFINE_PRIV_FROM_JS(GjsCairoContext, gjs_cairo_context_class);
 
 static void
@@ -251,19 +272,14 @@ _gjs_cairo_context_construct_internal(JSContext *context,
     priv->cr = cairo_reference(cr);
 }
 
-static JSBool
-gjs_cairo_context_constructor(JSContext *context,
-                              JSObject  *obj,
-                              uintN      argc,
-                              jsval     *argv,
-                              jsval     *retval)
+GJS_NATIVE_CONSTRUCTOR_DECLARE(cairo_context)
 {
+    GJS_NATIVE_CONSTRUCTOR_VARIABLES(cairo_context)
     JSObject *surface_wrapper;
     cairo_surface_t *surface;
     cairo_t *cr;
 
-    if (!gjs_check_constructing(context))
-        return JS_FALSE;
+    GJS_NATIVE_CONSTRUCTOR_PRELUDE(cairo_context);
 
     if (!gjs_parse_args(context, "Context", "o", argc, argv,
                         "surface", &surface_wrapper))
@@ -280,9 +296,11 @@ gjs_cairo_context_constructor(JSContext *context,
     if (!gjs_cairo_check_status(context, cairo_status(cr), "context"))
         return JS_FALSE;
 
-    _gjs_cairo_context_construct_internal(context, obj, cr);
+    _gjs_cairo_context_construct_internal(context, object, cr);
     cairo_destroy(cr);
 
+    GJS_NATIVE_CONSTRUCTOR_FINISH(cairo_context);
+
     return JS_TRUE;
 }
 
@@ -320,8 +338,8 @@ _GJS_CAIRO_CONTEXT_DEFINE_FUNC0(clipPreserve, cairo_clip_preserve)
 _GJS_CAIRO_CONTEXT_DEFINE_FUNC0AFFFF(clipExtents, cairo_clip_extents)
 _GJS_CAIRO_CONTEXT_DEFINE_FUNC0(closePath, cairo_close_path)
 _GJS_CAIRO_CONTEXT_DEFINE_FUNC0(copyPage, cairo_copy_page)
-_GJS_CAIRO_CONTEXT_DEFINE_FUNC0AFF(deviceToUser, cairo_device_to_user)
-_GJS_CAIRO_CONTEXT_DEFINE_FUNC0AFF(deviceToUserDistance, cairo_device_to_user_distance)
+_GJS_CAIRO_CONTEXT_DEFINE_FUNC2FFAFF(deviceToUser, cairo_device_to_user, "x", "y")
+_GJS_CAIRO_CONTEXT_DEFINE_FUNC2FFAFF(deviceToUserDistance, cairo_device_to_user_distance, "x", "y")
 _GJS_CAIRO_CONTEXT_DEFINE_FUNC0(fill, cairo_fill)
 _GJS_CAIRO_CONTEXT_DEFINE_FUNC0(fillPreserve, cairo_fill_preserve)
 _GJS_CAIRO_CONTEXT_DEFINE_FUNC0AFFFF(fillExtents, cairo_fill_extents)
@@ -380,8 +398,8 @@ _GJS_CAIRO_CONTEXT_DEFINE_FUNC0(stroke, cairo_stroke)
 _GJS_CAIRO_CONTEXT_DEFINE_FUNC0(strokePreserve, cairo_stroke_preserve)
 _GJS_CAIRO_CONTEXT_DEFINE_FUNC0AFFFF(strokeExtents, cairo_stroke_extents)
 _GJS_CAIRO_CONTEXT_DEFINE_FUNC2(translate, cairo_translate, "ff", double, tx, double, ty)
-_GJS_CAIRO_CONTEXT_DEFINE_FUNC0AFF(userToDevice, cairo_user_to_device)
-_GJS_CAIRO_CONTEXT_DEFINE_FUNC0AFF(userToDeviceDistance, cairo_user_to_device_distance)
+_GJS_CAIRO_CONTEXT_DEFINE_FUNC2FFAFF(userToDevice, cairo_user_to_device, "x", "y")
+_GJS_CAIRO_CONTEXT_DEFINE_FUNC2FFAFF(userToDeviceDistance, cairo_user_to_device_distance, "x", "y")
 
 
 static JSBool
@@ -516,6 +534,70 @@ maskSurface_func(JSContext *context,
     return JS_TRUE;
 }
 
+static JSBool
+setDash_func(JSContext *context,
+             uintN      argc,
+             jsval     *vp)
+{
+    jsval *argv = JS_ARGV(context, vp);
+    JSObject *obj = JS_THIS_OBJECT(context, vp);
+    guint i;
+    cairo_t *cr;
+    JSObject *dashes;
+    double offset;
+    JSBool retval = JS_FALSE;
+    guint len;
+    GArray *dashes_c = NULL;
+
+    if (!gjs_parse_args(context, "setDash", "of", argc, argv,
+                        "dashes", &dashes, "offset", &offset))
+        return JS_FALSE;
+
+    JS_AddObjectRoot(context, &dashes);
+
+    if (!JS_IsArrayObject(context, dashes)) {
+        gjs_throw(context, "dashes must be an array");
+        goto out;
+    }
+
+    if (!JS_GetArrayLength(context, dashes, &len)) {
+        gjs_throw(context, "Can't get length of dashes");
+        goto out;
+    }
+
+    dashes_c = g_array_sized_new (FALSE, FALSE, sizeof(double), len);
+    for (i = 0; i < len; ++i) {
+        jsval elem;
+        double b;
+
+        elem = JSVAL_VOID;
+        if (!JS_GetElement(context, dashes, i, &elem)) {
+            goto out;
+        }
+        if (elem == JSVAL_VOID)
+            continue;
+
+        if (!JS_ValueToNumber(context, elem, &b))
+            goto out;
+        if (b <= 0) {
+            gjs_throw(context, "Dash value must be positive");
+            goto out;
+        }
+
+        g_array_append_val(dashes_c, b);
+    }
+
+    cr = gjs_cairo_context_get_context(context, obj);
+    cairo_set_dash(cr, (double*)dashes_c->data, dashes_c->len, offset);
+    JS_SET_RVAL(context, vp, JSVAL_VOID);
+    retval = JS_TRUE;
+ out:
+    if (dashes_c != NULL)
+        g_array_free (dashes_c, TRUE);
+    JS_RemoveObjectRoot(context, &dashes);
+    return retval;
+}
+
 static JSBool
 setSource_func(JSContext *context,
                uintN      argc,
@@ -836,7 +918,7 @@ static JSFunctionSpec gjs_cairo_context_proto_funcs[] = {
     { "scale", (JSNative)scale_func, 0, JSFUN_FAST_NATIVE },
     { "selectFontFace", (JSNative)selectFontFace_func, 0, JSFUN_FAST_NATIVE },
     { "setAntialias", (JSNative)setAntialias_func, 0, JSFUN_FAST_NATIVE },
-    // setDash
+    { "setDash", (JSNative)setDash_func, 0, JSFUN_FAST_NATIVE },
     // setFontFace
     // setFontMatrix
     // setFontOptions
